from unittest import TestCase
from jscaffold.iounit.envvar import EnvVar
from jscaffold.iounit.iounit import Inputable
from jscaffold.widgets.widgetfactory import WidgetFactory


class TestWidgetFactory(TestCase):
    def test_widgetfactory_create_input_with_select(self):
        factory = WidgetFactory()
        var = EnvVar("VAR").select("A", "B", "C")
        input_widget = factory.create_input(var)
        assert input_widget.select_widget.value is None
        assert input_widget.select_widget.options == ("A", "B", "C")

        input_widget.value = "B"

        assert input_widget.select_widget.value == "B"
        assert input_widget.select_widget.options == ("A", "B", "C")

        input_widget.value = "value-non-existed"
        assert input_widget.select_widget.value is None
        assert input_widget.select_widget.options == ("A", "B", "C")

    def test_widgetfactory_create_upload_file_input(self):
        factory = WidgetFactory()
        var = EnvVar("VAR").upload_file()
        input_widget = factory.create_input(var)
        assert input_widget.type == "upload_file"

    def test_widgetfactory_create_non_Formatable_input(self):
        class DummyInput(Inputable):
            def _read(self, context=None):
                return "dummy"

        factory = WidgetFactory()
        input = DummyInput()
        factory.create_input(input)



from abc import ABC
import ipywidgets


class IPYWidgetWrapper(ABC):
    """
    IPython widget wrapper
    """

    def create_widget(self):
        raise NotImplementedError()

    def update_widget(self):
        raise NotImplementedError()

    def get_widget(self) -> "ipywidgets.Widget":
        raise NotImplementedError()

    def focus(self):
        widget = self.get_widget()

        if widget is not None:
            widget.focus()



import os

from jscaffold.iounit.format import Formatter
from .valuable import Valuable
from ..contexts.context import Context


class EnvVar(Valuable):
    """
    Wrapper for environment variable
    """

    def __init__(self, key):
        super().__init__()
        self._key = key

    def _get_key(self):
        return self._key

    def _get_id(self):
        return f"Env:{self.key}"

    def _write(self, value=None, context: Context = None):
        value = (
            Formatter(self.format, value).if_none_use_defaults().cast_to_format().value
        )
        if value is None:
            del os.environ[self.key]
        else:
            os.environ[self.key] = str(value)
        if context is not None and context.log is not None:
            context.log(f"Set {self.key}={self._format_display_value(value)}\n")

    def _read(self, context: Context = None):
        return Formatter(self.format, os.getenv(self.key)).cast_to_format().value



from unittest import TestCase
from unittest.mock import Mock, patch
from jscaffold.iounit.jsonfilevar import JsonFileVar
from tempfile import NamedTemporaryFile
import json
import textwrap

from jscaffold.iounit.valuable import Valuable


def write_to_tmp(dict):
    tmp_file = NamedTemporaryFile(delete=False)
    content = json.dumps(dict)
    tmp_file.write(bytes(content, "utf-8"))
    tmp_file.close()
    return tmp_file.name


@patch("jscaffold.services.changedispatcher.ChangeDispatcher.dispatch", Mock())
class TestJsonFileVar(TestCase):
    def setUp(self):
        Valuable.shared_storage = {}

    def test_jsonfilevar_get_id(self):
        variable = JsonFileVar("A", "./config.json")
        assert variable.id == "JsonFile:./config.json:A"

    def test_jsonfilevar_read_from_file(self):
        source = write_to_tmp({"A": {"B": {"C": "value"}}})
        variable = JsonFileVar("A.B.C", source)
        assert variable.value == "value"

    def test_jsonfilevar_read_boolean_from_file(self):
        source = write_to_tmp({"A": {"B": {"C": True}}})
        variable = JsonFileVar("A.B.C", source)
        assert variable.value is True

    def test_jsonfilevar_write_to_file(self):
        source = write_to_tmp({"A": {"B": {"C": "value"}}})
        variable = JsonFileVar("A.B.C", source).indent(4)
        variable.write("new_value")

        with open(source, "r") as file:
            content = file.read()
            assert content == textwrap.dedent(
                """\
            {
                "A": {
                    "B": {
                        "C": "new_value"
                    }
                }
            }"""
            )

    @patch("builtins.open")
    def test_jsonfilevar_str_reader(self, mock_open):
        json_str = '{"A": {"B": {"C": "value"}}}'
        mock_open.return_value.__enter__().read.return_value = json_str
        var = JsonFileVar("C", "mock-file").reader("A.B.C")
        assert var.value == "value"

    @patch("builtins.open")
    def test_jsonfilevar_func_reader(self, mock_open):
        json_str = '{"A": {"B": {"C": "value"}}}'
        mock_open.return_value.__enter__().read.return_value = json_str

        def query(dict):
            return dict["A"]["B"]["C"]

        var = JsonFileVar("C", "mock-file").reader(query)
        assert var.value == "value"

    @patch("builtins.open")
    def test_jsonfilevar_func_path(self, mock_open):
        json_str = '{"A": {"B": {"C": "value"}}}'
        mock_open.return_value.__enter__().read.return_value = json_str

        var = JsonFileVar("C", "mock-file").path("A.B.C")
        assert var.value == "value"

    @patch("builtins.open")
    def test_jsonfilevar_read_number(self, mock_open):
        json_str = '{"A": {"B": {"C": 1}}}'
        mock_open.return_value.__enter__().read.return_value = json_str

        var = JsonFileVar("C", "mock-file").path("A.B.C")
        assert var.value == 1

    @patch("builtins.open")
    def test_jsonfilevar_write_number(self, mock_open):
        json_str = '{"A": {"B": {"C": 1}}}'
        mock_open.return_value.__enter__().read.return_value = json_str
        var = JsonFileVar("C", "mock-file").path("A.B.C")
        var.update(2)
        mock_open.return_value.__enter__().write.assert_called_with(
            '{"A": {"B": {"C": 2}}}'
        )

    def test_use(self):
        with JsonFileVar.use("config.json"):
            var1 = JsonFileVar("A")
            assert var1.filename == "config.json"
            assert var1.state.indent is None

            var2 = JsonFileVar("B")
            assert var2.filename == "config.json"
            assert var2.state.indent is None

        assert JsonFileVar.default_filename is None
        assert JsonFileVar.default_indent is None

    def test_use_indent(self):
        with JsonFileVar.use("config.json").indent(9):
            var1 = JsonFileVar("A")
            assert var1.filename == "config.json"
            assert var1.state.indent == 9

            var2 = JsonFileVar("B")
            assert var2.filename == "config.json"
            assert var2.state.indent == 9

        assert JsonFileVar.default_filename is None
        assert JsonFileVar.default_indent is None

    def test_use_nested_indent(self):
        with JsonFileVar.use("config.json").indent(9):
            var1 = JsonFileVar("A")
            assert var1.filename == "config.json"
            assert var1.state.indent == 9

            with JsonFileVar.use("config2.json").indent(10):
                var2 = JsonFileVar("B")
                assert var2.filename == "config2.json"
                assert var2.state.indent == 10

            var3 = JsonFileVar("A")
            assert var3.filename == "config.json"
            assert var3.state.indent == 9



from unittest import IsolatedAsyncioTestCase
import uuid
from jscaffold.contexts.context import FormContext
from jscaffold.iounit.envfilevar import EnvFileVar
from jscaffold.iounit.envvar import EnvVar
from jscaffold.iounit.valuable import Valuable
from jscaffold.processor import Processor
from unittest.mock import MagicMock, Mock, patch
import asyncio
import pytest
from tempfile import NamedTemporaryFile
import os


@patch("jscaffold.services.changedispatcher.ChangeDispatcher.dispatch", Mock())
class TestProcessor(IsolatedAsyncioTestCase):
    def setUp(self):
        Valuable.shared_storage = {}

    @pytest.mark.asyncio()
    async def test_processor_execute_list(self):
        callback = MagicMock()
        var = EnvVar("VAR1")
        context = FormContext(input=var)
        processor = Processor(context)
        await processor(var, [callback, callback], "input")

        assert callback.call_count == 2

    @pytest.mark.asyncio()
    async def test_processor_run_func_pass_input(self):
        var = EnvVar("VAR1")
        context = FormContext(input=var)
        processor = Processor(context)
        result = None

        def callback(VAR1):
            nonlocal result
            result = VAR1

        await processor(var, callback, self.id())

        assert result == self.id()

    def test_processor_create_task(self):
        output = MagicMock()
        processor = Processor()
        done = MagicMock()
        task = processor.create_task(EnvVar("VAR1"), output, "input")
        task.add_done_callback(done)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(task)
        done.assert_called_once()

    def test_processor_run_script(self):
        output = """echo Hello"""
        context = MagicMock()
        processor = Processor(context)
        task = processor.create_task(None, output, None)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(task)
        context.log.assert_called_once_with("Hello\n")

    @pytest.mark.asyncio()
    async def test_processor_run_script_pass_variable(self):
        output = """echo $VAR1"""
        context = MagicMock()
        processor = Processor(context)
        tmp_file = NamedTemporaryFile(delete=True)

        var = EnvFileVar("VAR1", tmp_file.name)
        var.write("123")

        await processor(var, output, var.value)
        context.log.assert_called_once_with("123\n")

    @pytest.mark.asyncio()
    async def test_processor_run_script_pass_variable_from_value(self):
        """
        It should pass the value to the script instead of reading
        from input
        """
        os.environ["VAR1"] = ""
        output = """echo $VAR1"""
        context = MagicMock()
        processor = Processor(context)
        var = EnvVar("VAR1")
        await processor(var, output, "123")
        context.log.assert_called_once_with("123\n")

    @pytest.mark.asyncio()
    async def test_processor_raise_exception(self):
        """
        It should write the error message to context.log
        """

        def callback():
            raise Exception("Error")

        context = MagicMock()
        processor = Processor(context)
        await processor(None, callback, context)
        message = context.log.call_args[0][0]

        assert message.split("\n")[0] == "Traceback (most recent call last):"

    @pytest.mark.asyncio()
    async def test_processor_raise_exception_should_stop(self):
        def stop():
            raise Exception("Error")

        callback = MagicMock()
        context = MagicMock()
        processor = Processor(context)
        await processor(None, [stop, callback], context)
        message = context.log.call_args[0][0]

        assert message.split("\n")[0] == "Traceback (most recent call last):"
        assert callback.call_count == 0

    @pytest.mark.asyncio
    async def test_processor_add_apply_to_source(self):
        """
        It should add the apply method to the source
        if save_changes was set in context
        """
        key = str(uuid.uuid4())
        var = EnvVar(key)
        context = MagicMock()
        context.save_changes = False
        context.input = var
        processor = Processor(context)
        await processor(var, None, "123")
        assert os.environ.get(key) is None

        context.save_changes = True
        processor = Processor(context)
        await processor(var, None, "123")
        assert os.environ.get(key) == "123"

    def test_processor_invoke(self):
        input = EnvVar("VAR1")
        context = FormContext(input=input)
        process = Processor(context)
        is_called = False

        def callback(log, value, VAR1):
            nonlocal is_called
            is_called = True
            assert value == "value"
            assert VAR1 == "value"
            assert log == context.log

        process.invoke(callback, "value", context)
        assert is_called is True



from jscaffold.utils import inspect_arg_name


def test_inspect_arg_name_by_kwarg():
    def func(a, b, c):
        return inspect_arg_name(2, "c")

    A = 1
    C = 4
    res = func(A, c=C, b="3")

    assert res == "C"


def test_inspect_arg_name_by_pos():
    def func(a, b, c):
        return inspect_arg_name(2, "c")

    A = 1
    D = 4
    res = func(A, "b", D)

    assert res == "D"


def test_inspect_arg_name_not_found():
    def func(a, b, c):
        return inspect_arg_name(4, "k")

    A = 1
    D = 4
    res = func(A, "b", D)
    assert res is None


def test_inspect_arg_name_not_varible():
    def func(a, b, c):
        return inspect_arg_name(0, "a")

    D = 4
    res = func("1===2", "b", D)
    assert res is None


def test_inspect_arg_name_can_not_accpt_func():
    def func(a, b, c):
        return inspect_arg_name(0, "a")

    def func2():
        return "a"

    D = 4
    res = func(func2(), "b", D)
    assert res is None



class SharedStorage:
    """
    A key-value based shared storage between Jupyter cells
    """

    def __init__(self):
        self.data = {}

    def set(self, key, value):
        self.data[key] = value

    def get(self, key):
        return self.data.get(key)

    def remove(self, key):
        if key in self.data:
            del self.data[key]


shared_storage = SharedStorage()



import asyncio


class KeyFilterDebouncer:
    def __init__(self, delay):
        """Initializes the Debouncer with a specific delay in seconds."""
        self.delay = delay
        self.timer_tasks = {}
        self.tasks = {}
        self.task_id = 0

    def debounce(self, key, func, *args, **kwargs):
        async def run():
            if key in self.timer_tasks:
                self.timer_tasks[key].cancel()  # Cancel the previous task
            try:
                self.timer_tasks[key] = asyncio.get_event_loop().create_task(
                    asyncio.sleep(self.delay)
                )
                await self.timer_tasks[key]
                del self.timer_tasks[key]
                func(*args, **kwargs)
            except asyncio.CancelledError:
                pass

        asyncio.ensure_future(run())

    def __call__(self, key, func, *args, **kwargs):
        self.debounce(key, func, *args, **kwargs)



from contextlib import contextmanager
from jscaffold.patchers.assign import PatchAssignment
from .valuable import Valuable


class EnvFileVar(Valuable):
    default_filename = None

    def __init__(self, key, filename=None):
        super().__init__()

        if filename is not None:
            self.filename = filename
        elif EnvFileVar.default_filename is not None:
            self.filename = EnvFileVar.default_filename
        else:
            raise ValueError("filename is not provided")
        self._key = key
        self.patcher = PatchAssignment()

    def _get_key(self):
        return self._key

    def _get_id(self):
        return f"EnvFile:{self.filename}:{self.key}"

    def _write(self, value, context=None):
        content = self._read_file_content()
        replaced, _ = self.patcher(
            content if content is not None else "", self.key, value
        )

        file = open(self.filename, "w")
        file.write(replaced)
        file.close()

        if context is not None and context.log is not None:
            context.log(
                f"Set {self.key}={self._format_display_value(value)} to {self.filename}\n"
            )

    def _read(self, context=None):
        content = self._read_file_content()
        if content is None:
            return None

        _, value = self.patcher(content, self.key)
        return value

    def _read_file_content(self):
        try:
            file = open(self.filename, "r")
            content = file.read()
            file.close()
            self.content = content
        except FileNotFoundError:
            return None
        return content

    @classmethod
    @contextmanager
    def use(cls, filename: str):
        try:
            EnvFileVar.default_filename = filename
            yield
        finally:
            EnvFileVar.default_filename = None



from unittest import TestCase
from jscaffold.form import form
from jscaffold.iounit.var import Var


class TestForm(TestCase):
    def test_form_normalize_str(self):
        """
        Form can be created
        """
        created_form = form("VAR")
        assert created_form is not None
        assert isinstance(created_form.input[0], Var) is True

    def test_nested_form(self):
        form("VAR1").form("VAR2")



from jscaffold.services.sharedstorage import shared_storage


class Context:
    def __init__(
        self,
        main_panel=None,
        shared_storage=shared_storage,
        log_view=None,
    ):
        # Shared storage between input and output
        self.shared_storage = shared_storage
        self.log_view = log_view
        self.main_panel = main_panel

    def to_kwargs(self):
        return {
            "log": self.log,
            "clear_log": self.clear_log,
            "shared_storage": self.shared_storage,
        }

    def log(self, *args):
        if self.log_view is not None:
            self.log_view.append_stdout(*args)
        else:
            print(*args)

    def clear_log(self):
        if self.log_view is not None:
            self.log_view.clear_output()

    def on_start_processing(self):
        if self.log_view is not None:
            self.log_view.is_running = True

    def on_stop_processing(self):
        if self.log_view is not None:
            self.log_view.is_running = False


class FormContext(Context):
    def __init__(self, input=None, save_changes=True, parent=None, form=None, **kwargs):
        super().__init__(**kwargs)
        self.input = input
        self.save_changes = save_changes
        self.parent = parent
        self.form = form

    @staticmethod
    def from_base_context(base_context, **kwargs):
        return FormContext(
            log_view=base_context.log_view,
            shared_storage=base_context.shared_storage,
            main_panel=base_context.main_panel,
            parent=base_context,
            **kwargs
        )

    def refresh_form(self):
        if self.form is not None:
            self.form.refresh()

    def to_kwargs(self):
        kwargs = super().to_kwargs()
        kwargs.update(
            {
                "input": self.input,
                "refresh_form": self.refresh_form,
            }
        )
        return kwargs



import anywidget
import traitlets
import pathlib


class LogViewWidget(anywidget.AnyWidget):
    _esm = pathlib.Path(__file__).parent / "index.js"
    _css = pathlib.Path(__file__).parent / "style.css"

    value = traitlets.Unicode("").tag(sync=True)
    is_running = traitlets.Bool(False).tag(sync=True)

    def append_stdout(self, *args):
        message = " ".join(str(arg) for arg in args)
        self.value = self.value + message

    def clear_output(self):
        self.value = ""



import uuid
from jscaffold.iounit.envvar import EnvVar
from jscaffold.widgets.widgetfactory import WidgetFactory
from jscaffold.widgets.inputwidget import (
    TextInputWidget,
    TextAreaInputWidget,
    SelectInputWidget,
    FileUploadInputWidget,
    LocalPathInputWidget,
)


def test_textinputwidget_placeholder():
    factory = WidgetFactory()

    var = EnvVar(str(uuid.uuid4()))
    input_widget = factory.create_input(var)
    assert input_widget.text_widget.placeholder == ""

    var = EnvVar(str(uuid.uuid4())).defaults("placeholder")
    input_widget = factory.create_input(var)
    assert input_widget.text_widget.placeholder == "placeholder"


def test_create():
    var = EnvVar("var").select([])
    widgets = [
        TextInputWidget,
        TextAreaInputWidget,
        SelectInputWidget,
        FileUploadInputWidget,
        LocalPathInputWidget,
    ]
    for widget in widgets:
        widget(var)


def test_observe():
    var = EnvVar("var").select([])
    widgets = [
        TextInputWidget,
        TextAreaInputWidget,
        SelectInputWidget,
        FileUploadInputWidget,
        LocalPathInputWidget,
    ]
    for widget in widgets:
        w = widget(var)
        w.observe(lambda x: x)



from jscaffold.debounce import KeyFilterDebouncer


class ChangeDispatcher:
    def __init__(self):
        # It can't use WeakSet because it will be removed in Jupyter environment
        self.listeners = set()
        self.queue = []
        self.debouncer = KeyFilterDebouncer(0.1)

    def add_listener(self, listener):
        self.listeners.add(listener)

    def dispatch(self, key, payload):
        payload_type = (
            payload["type"] if payload is not None and "type" in payload else None
        )

        debounce_key = f"{key}:{payload_type}" if payload_type is not None else key

        self.debouncer(debounce_key, self._invoke_listeners, key, payload)

    def _invoke_listeners(self, key, payload):
        pending_remove_list = []
        for listener in self.listeners:
            try:
                listener(key, payload)
            except Exception:
                pending_remove_list.append(listener)

        for listener in pending_remove_list:
            self.listeners.remove(listener)


class Listener:
    def __init__(self, type, callback):
        self.type = type
        self.callback = callback

    def __call__(self, type, payload):
        if self.type == type:
            self.callback(payload)


change_dispatcher = ChangeDispatcher()



from jscaffold.task.runtask import RunTask
from jscaffold.iounit.applytosource import ApplyToSource
import asyncio
from inspect import signature
import traceback


class Processor:
    """
    Run output(s) from input(s) and instance value(s)
    """

    def __init__(self, context=None):
        self.context = context

    async def __call__(self, input, runnables, value):
        return await self.process(input, runnables, value)

    # pylama:ignore=C901
    async def process(self, input, runnables, value):
        context = self.context
        if context is not None:
            context.clear_log()

        runnables = [runnables] if not isinstance(runnables, list) else runnables[:]

        if self.context is not None and self.context.save_changes is True:
            runnables.insert(0, ApplyToSource())

        if input is None:
            inputs = []
            values = []
        else:
            inputs = [input] if not isinstance(input, list) else input
            values = [value] if not isinstance(value, list) else value

        env = dict(
            [
                (
                    i.key,
                    str(v) if v is not None else "",
                )
                for (i, v) in zip(inputs, values)
            ]
        )

        if context is not None:
            context.on_start_processing()

        for target in runnables:
            try:
                if isinstance(target, str):
                    script = target
                    run_task = RunTask()
                    run_task.script = script
                    await run_task(print=self.context.log, env=env)
                elif callable(target):
                    self.invoke(target, value, self.context)
            except Exception:
                if self.context is not None:
                    self.context.log(traceback.format_exc())
                # Stop processing if an error occurs
                break

        if context is not None:
            context.on_stop_processing()

    def create_task(self, input, output, value):
        async def run():
            return await self.process(input, output, value)

        return asyncio.get_event_loop().create_task(run())

    def invoke(self, callable, value, context):
        sig = signature(callable)
        args = {}
        context_kwargs = context.to_kwargs()
        for key in context_kwargs:
            if key in sig.parameters:
                args[key] = context_kwargs[key]
        if "value" in sig.parameters:
            args["value"] = value
        if "context" in sig.parameters:
            args["context"] = context
        input = context.input if isinstance(context.input, list) else [context.input]
        value = value if isinstance(value, list) else [value]
        for index, i in enumerate(input):
            if i.key in sig.parameters and i.key not in args:
                args[i.key] = value[index]
        return callable(**args)



from jscaffold import Var


def test_valuable_should_share_value(request):
    key = request.node.name
    var1 = Var(key)
    var2 = Var(key)

    var1.update("test")
    assert var2.value == "test"

    var1.update("test2")
    assert var2.value == "test2"



from unittest.mock import MagicMock
from jscaffold.decorators.iot import preset_iot
from jscaffold.iounit.var import Var


def test_iot_preset_should_change_str_to_shared_storage_var():
    res = MagicMock()

    @preset_iot
    def func(
        input=None,
        output=None,
        title=None,
    ):
        res(input)

    func("input")

    inputs = res.call_args_list[0][0]
    input = inputs[0]
    assert isinstance(input, Var) is True
    assert input.key == "input"



from .iounit import Inputable, Outputable


class ApplyToSource(Outputable):
    """
    An Outputable that writes the value to the source input.
    """

    def __call__(self, value, context):
        values = []
        sources = []
        if isinstance(value, list):
            values = value
            sources = context.input
        else:
            values = [value]
            sources = [context.input]

        for index, value in enumerate(values):
            source = sources[index]
            if isinstance(source, Inputable):
                if source.format.readonly:
                    continue
            if isinstance(source, Outputable):
                source.write(value, context)



from unittest import TestCase
from jscaffold.iounit.envvar import EnvVar
import os
from unittest.mock import Mock, patch


@patch("jscaffold.services.changedispatcher.ChangeDispatcher.dispatch", Mock())
class TestEnvVar(TestCase):
    def test_envvar_get_id(self):
        var = EnvVar("VALUE_NOT_EXISTED")
        assert var._get_id() == "Env:VALUE_NOT_EXISTED"

    def test_envvar_defaults_is_array(self):
        var = EnvVar("VALUE_NOT_EXISTED").defaults(["V1", "V2"])

        assert var.value == "V1"

    def test_envvar_write_without_value(self):
        name = "7ccb192e-4de4-4720-aa65-e3e687e7a5eb"
        var = EnvVar(name).defaults("default")
        var()
        assert os.getenv(name) == "default"

    def test_envvar_write_none(self):
        var = EnvVar("var1")
        var.write("test")
        var.write(None)

        assert os.getenv("var1", None) is None

    def test_envvar_write_integer(self):
        var = EnvVar("var1")
        var.write(123)
        assert os.getenv("var1") == "123"

    @patch("jscaffold.services.changedispatcher.change_dispatcher.dispatch")
    def test_envvar_write_should_dispatch_change(self, mock_dispatch):
        var = EnvVar("VAR")
        var.write("value")

        mock_dispatch.assert_called_once_with(
            "Env:VAR", {"type": "value_changed", "value": "value"}
        )

    def test_envvar_read_integer(self):
        os.environ["var1"] = "123"
        var = EnvVar("var1").number()
        assert var.value == 123



from jscaffold.patchers.dict import PatchDict
from .valuable import Valuable
from collections import OrderedDict
import json


class JsonFileVar(Valuable):
    default_filename = None
    default_indent = None

    class State:
        def __init__(self):
            self.key = None
            self.indent = None
            self.reader = None
            self.path = None

    def __init__(self, key, filename=None):
        super().__init__()
        if filename is not None:
            self.filename = filename
        elif JsonFileVar.default_filename is not None:
            self.filename = JsonFileVar.default_filename
        else:
            raise ValueError("filename is not provided")

        self.patcher = PatchDict()
        self.state = JsonFileVar.State()
        self.state.path = key
        self.state.key = key
        self.state.indent = JsonFileVar.default_indent

    def indent(self, indent):
        self.state.indent = indent
        return self

    def _get_key(self):
        return self.state.key

    def _get_id(self):
        return f"JsonFile:{self.filename}:{self.state.key}"

    def reader(self, value):
        self.state.reader = value
        return self

    def path(self, value):
        self.state.path = value
        return self

    def _write(self, value, context=None):
        content = self._read_json_from_file()
        if content is None:
            content = {}
        new_content = self.patcher.write(content, self.state.path, value)
        with open(self.filename, "w") as file:
            file.write(json.dumps(new_content, indent=self.state.indent))
            file.close()

        if context is not None and context.log is not None:
            context.log(
                f"Set {self.key}={self._format_display_value(value)} to {self.filename}\n"
            )

    def _read(self, context=None):
        content = self._read_json_from_file()
        if content is None:
            return None

        if callable(self.state.reader):
            value = self.state.reader(content)
        else:
            value = self.patcher.read(
                content,
                self.state.reader if self.state.reader is not None else self.state.path,
            )
        if value is None:
            return None
        return value

    def _read_json_from_file(self):
        try:
            with open(self.filename, "r") as file:
                content = file.read()
            json_content = json.loads(content, object_pairs_hook=OrderedDict)
            return json_content
        except FileNotFoundError:
            return None

    @classmethod
    def use(cls, filename: str):
        class JsonFileVarContext:
            def __init__(self, filename):
                self.prev_filename = None
                self.prev_indent = None
                self.current_filename = filename
                self.current_indent = None

            def __enter__(self):
                self.prev_filename = JsonFileVar.default_filename
                self.prev_indent = JsonFileVar.default_indent
                JsonFileVar.default_filename = self.current_filename
                JsonFileVar.default_indent = self.current_indent
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                JsonFileVar.default_filename = self.prev_filename
                JsonFileVar.default_indent = self.prev_indent

            def indent(self, indent):
                self.current_indent = indent
                return self

        return JsonFileVarContext(filename)



from ipywidgets import widgets
from jscaffold.iounit.format import FileSource, FormatType
from jscaffold.widgets.inputwidget import (
    LocalPathInputWidget,
    SelectInputWidget,
    FileUploadInputWidget,
    TextAreaInputWidget,
    TextInputWidget,
    NumberInputWidget,
    InputWidgetType,
)
from ..iounit.iounit import Inputable


class WidgetWrapper:
    def __init__(
        self,
        widget,
        container,
        get_value=None,
        on_click=None,
        set_value=None,
        type=None,
    ):
        self.widget = widget
        self.container = container
        self.get_value = get_value
        self.set_value = set_value
        self.on_click = on_click
        self.type = type

    def focus(self):
        self.widget.focus()


class WidgetFactory:
    def create_link_button(self, description):
        style = """
        <style>
        .jsacffold-link-button {
            background: transparent;
            color: #1976d2;
            width: auto;
        }

        .jsacffold-link-button:hover {
            background: transparent;
            border: none;
            box-shadow: none ! important;
            opacity: 0.5;
        }

        .jsacffold-link-button:active {
            background: transparent;
            border: none;
            box-shadow: none ! important;
            opacity: 0.7;
        }

        .jsacffold-link-button:focus {
            border: none;
            box-shadow: none ! important;
            outline: none ! important;
        }
        </style>
        """
        style_html = widgets.HTML(style)

        button = widgets.Button(description=description)
        button.add_class("jscaffold-link-button")
        container = widgets.Box([button, style_html])

        return WidgetWrapper(
            button, container, on_click=lambda callback: button.on_click(callback)
        )

    def get_input_widget_type(self, input: Inputable):
        if not hasattr(input, "format"):
            return InputWidgetType.Text
        format = input.format
        if format.type == FormatType.Number.value:
            return InputWidgetType.Number
        elif (
            format.type == FormatType.File.value
            and format.file_source == FileSource.Upload.value
        ):
            return InputWidgetType.FileUpload
        if (
            format.type == FormatType.File.value
            and format.file_source == FileSource.Local.value
        ):
            return InputWidgetType.LocalPath
        if input is not None and isinstance(format.select, list):
            return InputWidgetType.Select
        if format.multiline is True or (
            isinstance(format.multiline, int) and format.multiline > 1
        ):
            return InputWidgetType.Textarea
        return InputWidgetType.Text

    def create_input(self, input: Inputable):
        input_widget_type = self.get_input_widget_type(input)

        if input_widget_type == InputWidgetType.Number:
            return NumberInputWidget(input)
        elif input_widget_type == InputWidgetType.Select:
            return SelectInputWidget(input)
        elif input_widget_type == InputWidgetType.Textarea:
            return TextAreaInputWidget(input)
        elif input_widget_type == InputWidgetType.FileUpload:
            return FileUploadInputWidget(input)
        elif input_widget_type == InputWidgetType.LocalPath:
            return LocalPathInputWidget(input)

        return TextInputWidget(input)

    def create_submit_area(self, on_submit, default_label="Submit"):
        submit_button = widgets.Button(description=default_label)

        def button_callback(_):
            on_submit()

        submit_button.on_click(button_callback)

        return (submit_button, submit_button)  # hbox, confirm_button



from jscaffold.iounit.format import Formatter
from .valuable import Valuable
from ..contexts.context import Context
from jscaffold.services.sharedstorage import SharedStorage, shared_storage


class Var(Valuable):
    def __init__(self, key, shared_storage: SharedStorage = shared_storage):
        super().__init__()
        self._key = key
        self._shared_storage = shared_storage

    def _get_key(self):
        return self._key

    def _get_id(self):
        return f"Var:{self.key}"

    def _write(self, value=None, context: Context = None):
        formatter = Formatter(self.format, value)
        validaed_value = formatter.if_none_use_defaults().value
        self._shared_storage.set(self.key, validaed_value)
        if context is not None and context.log is not None:
            context.log(f"Set {self.key}={value}\n")

    def _read(self, context: Context = None):
        return self._shared_storage.get(self.key)



from enum import Enum
from jscaffold.iounit.format import Format
from jscaffold.iounit.iounit import Inputable
from jscaffold.services.tkservice import tk_serivce
from ipywidgets import widgets
from jscaffold.iounit.valuable import Valuable
import tempfile
import os
from pathlib import Path

from jscaffold.widgets.ipywidgetwrapper import IPYWidgetWrapper


def read_input(input: Inputable):
    if isinstance(input, Valuable):
        return input.value
    else:
        return input.read()


class InputWidgetType(Enum):
    Text = "text"
    Select = "select"
    Textarea = "textarea"
    FileUpload = "upload_file"
    LocalPath = "local_path"
    Number = "number"


class InputWidget(IPYWidgetWrapper):
    def __init__(self, type, input, child):
        self.type = type
        self.input = input
        format = input.format if hasattr(input, "format") else Format()
        self.desc_html = widgets.HTML(
            value=format.desc if format.desc is not None else ""
        )
        self.widget = widgets.VBox([child, self.desc_html])
        self.update_widget()

    def update_widget(self):
        format = self.input.format if hasattr(self.input, "format") else Format()
        desc = format.desc
        if desc is not None:
            self.desc_html.value = desc
        self.desc_html.layout.visibility = "visible" if desc else "hidden"

    def get_widget(self):
        return self.widget

    @property
    def value(self):
        raise NotImplementedError

    @value.setter
    def value(self, _value):
        raise NotImplementedError

    def observe(self, func):
        raise NotImplementedError


class TextInputWidget(InputWidget):
    def __init__(self, input: Inputable):
        value = input.read() if input is not None else None
        format = input.format if hasattr(input, "format") else Format()
        password = format.password
        placeholder = input.get_defaults() if hasattr(input, "get_defaults") else None
        if placeholder is None:
            placeholder = ""
        layout = widgets.Layout(width="360px")
        if password is True:
            text_widget = widgets.Password(
                value=value,
                layout=layout,
                placeholder=placeholder,
                disabled=format.readonly,
            )
        else:
            text_widget = widgets.Text(
                value=value,
                layout=layout,
                placeholder=placeholder,
                disabled=format.readonly,
            )

        self.text_widget = text_widget
        super().__init__(InputWidgetType.Text.value, input, text_widget)

    @property
    def value(self):
        return self.text_widget.value

    @value.setter
    def value(self, value):
        if value is None:
            value = ""
        if self.text_widget.value != value:
            self.text_widget.value = str(value) if value is not None else ""

    def observe(self, func):
        self.text_widget.observe(func)


DEFAULT_ROW_COUNT = 5


class TextAreaInputWidget(InputWidget):
    def __init__(self, input: Inputable):
        value = input.read()
        placeholder = input.get_defaults()
        if placeholder is None:
            placeholder = ""
        format = input.format
        rows = (
            format.multiline
            if not isinstance(format.multiline, bool)
            else DEFAULT_ROW_COUNT
        )
        layout = widgets.Layout(width="360px")
        textarea = widgets.Textarea(
            value=value,
            rows=rows,
            placeholder=placeholder,
            layout=layout,
            disabled=format.readonly,
        )
        self.textarea = textarea
        super().__init__(InputWidgetType.Textarea.value, input, textarea)

    @property
    def value(self):
        return self.textarea.value

    @value.setter
    def value(self, value):
        if value is None:
            value = ""
        if self.textarea.value != value:
            self.textarea.value = str(value) if value is not None else ""

    def observe(self, func):
        return self.textarea.observe(func)


class SelectInputWidget(InputWidget):
    def __init__(self, input: Inputable):
        value = input.read()
        format = input.format
        self.format = format
        if value not in format.select:
            if input.defaults in format.select:
                value = input.defaults
            else:
                value = None
        layout = widgets.Layout(width="360px")
        select_widget = widgets.Select(
            options=format.select, value=value, disabled=format.readonly, layout=layout
        )
        self.select_widget = select_widget
        super().__init__(InputWidgetType.Select.value, input, select_widget)

    @property
    def value(self):
        return self.select_widget.value

    @value.setter
    def value(self, value):
        if value not in self.format.select:
            value = None
        self.select_widget.value = value

    def observe(self, func):
        return self.select_widget.observe(func)

    def update_widget(self):
        super().update_widget()
        self.select_widget.disabled = self.format.readonly
        orig_value = self.select_widget.value
        self.select_widget.options = self.format.select
        if orig_value in self.format.select:
            self.select_widget.value = orig_value


class FileUploadInputWidget(InputWidget):
    def __init__(self, input: Inputable):
        value = str(input) if input is not None else None

        format = input.format
        text_box = widgets.Text(value=value, disabled=format.readonly)
        uploader = widgets.FileUpload(multiple=False)
        self.text_box = text_box

        def get_upload_folder():
            if format.upload_folder is not None:
                return format.upload_folder
            base_dir = tempfile.mkdtemp()
            if format.mkdir:
                Path(base_dir).mkdir(parents=True, exist_ok=True)
            return base_dir

        def on_upload_change(change):
            if change["type"] == "change" and change["name"] == "value":
                (file_dict,) = uploader.value
                # Remarks: type, content, size inside the file_dict
                filename = file_dict["name"]
                base_dir = get_upload_folder()
                abs_path = os.path.join(base_dir, filename)
                named_file = open(abs_path, "wb")
                named_file.write(file_dict["content"])
                named_file.close()
                text_box.value = abs_path

        uploader.observe(on_upload_change, names="value")

        hbox = widgets.HBox([text_box, uploader], layout=widgets.Layout(width="360px"))
        super().__init__(InputWidgetType.FileUpload.value, input, hbox)

    @property
    def value(self):
        return self.text_box.value

    @value.setter
    def value(self, value):
        if self.text_box.value != value:
            self.text_box.value = str(value) if value is not None else ""

    def observe(self, func):
        return self.text_box.observe(func)


class LocalPathInputWidget(InputWidget):
    def __init__(self, input: Inputable):
        def on_click(_):
            self.browser_button.disabled = True
            file_path = tk_serivce.open_file_dialog(input.format.file_type)
            self.browser_button.disabled = False
            if file_path == "":
                return
            text_box.value = file_path

        value = input.read()
        format = input.format
        placeholder = input.get_defaults()
        if placeholder is None:
            placeholder = ""
        text_box = widgets.Text(
            value=value,
            placeholder=placeholder,
            disabled=format.readonly,
        )
        browse_button = widgets.Button(description="Browse", disabled=format.readonly)
        self.text_box = text_box
        self.browser_button = browse_button
        hbox = widgets.HBox(
            [text_box, browse_button], layout=widgets.Layout(width="360px")
        )
        browse_button.on_click(on_click)
        super().__init__(InputWidgetType.LocalPath.value, input, hbox)

    @property
    def value(self):
        return self.text_box.value

    @value.setter
    def value(self, value):
        if self.text_box.value != value:
            self.text_box.value = str(value) if value is not None else ""

    def observe(self, func):
        return self.text_box.observe(func)


class NumberInputWidget(InputWidget):
    def __init__(self, input: Inputable):
        value = read_input(input)
        placeholder = input.get_defaults()
        if placeholder is None:
            placeholder = ""
        format = input.format
        layout = widgets.Layout(width="360px")

        float_text = widgets.FloatText(
            value=value,
            plreaceholder=placeholder,
            layout=layout,
            disabled=format.readonly,
        )

        self.float_text = float_text
        super().__init__(InputWidgetType.Textarea.value, input, float_text)

    def update_widget(self):
        super().update_widget()
        format = self.input.format
        self.float_text.disabled = format.readonly

    @property
    def value(self):
        return self.float_text.value

    @value.setter
    def value(self, value):
        if value is not None and self.float_text.value != value:
            self.float_text.value = float(value)

    def observe(self, func):
        return self.float_text.observe(func)



from contextvars import Context
from .iounit import IOAble
from .format import Format, Formatable, Formatter
import copy


class Valuable(IOAble, Formatable):
    shared_storage = {}

    def __init__(self):
        self.format = Format()

    def copy(self):
        return copy.deepcopy(self)

    def write(self, value=None, context: Context = None):
        super().write(value, context=context)
        id = self._get_id()
        Valuable.shared_storage[id] = value

    def refresh(self, context=None):
        self._has_cached_value = True
        latest = self._read(context=context)
        latest = Formatter(self.format, latest).if_none_use_defaults().value
        id = self._get_id()
        Valuable.shared_storage[id] = latest
        return self

    def update(self, value, context=None):
        self.write(value, context=context)
        id = self._get_id()
        Valuable.shared_storage[id] = value
        return self

    @property
    def value(self):
        # value should not be writable
        id = self._get_id()
        if id not in Valuable.shared_storage:
            self.refresh()
        return Valuable.shared_storage[id] if id in Valuable.shared_storage else None



from unittest.mock import MagicMock
from jscaffold.task.runtask import RunTask
import pytest
import threading


@pytest.mark.asyncio()
async def test_runtask_should_print_at_main_thread():
    content = []
    is_main_thread = None

    def print(line):
        content.append(line)
        nonlocal is_main_thread
        is_main_thread = threading.current_thread() is threading.main_thread()

    script = """\
    echo 1
    echo 2
    echo 3
    """

    runtask = RunTask()
    runtask.script = script
    await runtask(print=print)

    assert content == ["1\n", "2\n", "3\n"]
    assert is_main_thread is True


@pytest.mark.asyncio()
async def test_runtask_should_support_env():
    env = {"JS_VALUE": "test"}
    script = "echo $JS_VALUE"
    print = MagicMock()

    runtask = RunTask()
    runtask.script = script
    await runtask(print=print, env=env)
    print.assert_called_once_with("test\n")



from jscaffold.patchers.block import PatchBlock
import textwrap


def test_block_patch_read():
    source = textwrap.dedent(
        """\
        BEGIN
        value
        END
        """
    )

    patcher = PatchBlock("BEGIN", "END")

    assert patcher.read(source) == "value"


def test_block_patch_read_same_divider():
    source = textwrap.dedent(
        """\
        SEPARATOR
        value
        SEPARATOR
        """
    )

    patcher = PatchBlock("SEPARATOR", "SEPARATOR")

    assert patcher.read(source) == "value"


def test_block_patch_read_not_existed():
    source = textwrap.dedent(
        """\
        BEGIN
        value
        END
        """
    )

    patcher = PatchBlock("BEGIN", "BEGIN")

    assert patcher.read(source) is None


def test_block_patch_write():
    source = textwrap.dedent(
        """\
        BEGIN
        value
        END
        """
    )

    expected = textwrap.dedent(
        """\
        BEGIN
        value2
        END
        """
    )

    patcher = PatchBlock("BEGIN", "END")

    assert patcher.write(source, "value2") == expected


def test_block_write_non_existed():
    source = textwrap.dedent(
        """\
        """
    )

    expected = textwrap.dedent(
        """\

        BEGIN
        value2
        END"""
    )

    patcher = PatchBlock("BEGIN", "END")

    assert patcher.write(source, "value2") == expected



from jscaffold.utils import args_to_list
from jscaffold.widgets.ipywidgetwrapper import IPYWidgetWrapper
from ..contexts.context import FormContext
from IPython.display import display
from jscaffold.views.logview.logview import LogViewWidget
from ipywidgets import widgets
from jscaffold.widgets.widgetfactory import WidgetFactory
from jscaffold.processor import Processor
from jscaffold.services.changedispatcher import (
    Listener,
    change_dispatcher,
)
from jscaffold.debounce import KeyFilterDebouncer
from jscaffold.iounit.var import Var


class FormPanel(IPYWidgetWrapper):
    class State:
        """
        The State object hold the properties
        that may need to refresh the UI when
        changed
        """

        def __init__(self):
            self.instant_update = False
            self.title = None
            self.action_label = "Submit"
            self.save_changes = True
            self.runnables = []
            self.is_submitting = False

    def __init__(
        self,
        input=None,
        log_view=None,
        context=None,
    ):
        if input is None:
            self.input = []
        else:
            self.input = input if isinstance(input, list) else [input]
            self.input = list(
                map(lambda x: Var(x) if isinstance(x, str) else x, self.input)
            )

        self.widget = None
        self.state = FormPanel.State()
        self.submit_debouncer = KeyFilterDebouncer(0.1)

        has_own_log_view = False
        if log_view is None:
            has_own_log_view = True
            log_view = LogViewWidget()
        self.log_view = log_view

        if context is None:
            self.context = FormContext(
                input=self.input, log_view=self.log_view, form=self
            )
        else:
            self.context = FormContext.from_base_context(
                context, input=self.input, form=self
            )

        self.create_widget(has_own_log_view)

    def create_widget(self, has_own_log_view):
        factory = WidgetFactory()
        input = self.input

        items = []
        # TODO: Update title style
        self.title_widget = widgets.HTML(value=self.state.title)
        items.append(self.title_widget)

        self.input_widgets = []

        if len(input) > 0:
            grid = widgets.GridspecLayout(len(self.input), 2)

            grid._grid_template_columns = (
                "auto 1fr"  # A dirty hack to override the default value
            )
            grid._grid_template_rows = "auto"

            for i, item in enumerate(self.input):
                label = widgets.Label(value=item.key, layout=widgets.Layout())
                label.layout.margin = "0px 20px 0px 0px"

                input_widget = factory.create_input(item)
                grid[i, 0] = label
                grid[i, 1] = input_widget.widget
                self.input_widgets.append(input_widget)

            items.append(grid)

        (submit_area, confirm_button) = factory.create_submit_area(
            on_submit=lambda: self.submit(),
            default_label=self.state.action_label,
        )
        self.confirm_button = confirm_button
        self.submit_area = submit_area

        items.append(submit_area)
        if has_own_log_view:
            items.append(self.log_view)

        self.widget = widgets.VBox(items)
        self.listen()

    def get_widget(self):
        return self.widget

    def listen(self):
        def on_user_change(change):
            if (
                change["type"] == "change"
                and change["name"] == "value"
                and self.state.instant_update is True
            ):
                self.submit()

        # pylama:ignore=C901
        def create_listener(id, widget):
            def on_change(payload):
                value = payload["value"]
                type = payload["type"]
                if type == "value_changed":
                    try:
                        if widget.value != value:
                            widget.value = value
                    except Exception as e:
                        self.context.log(str(e))
                        raise e
                if type == "format_changed":
                    widget.update_widget()

            listener = Listener(id, on_change)
            return listener

        for input_widget, input_item in zip(self.input_widgets, self.input):
            listener = create_listener(input_item._get_id(), input_widget)
            change_dispatcher.add_listener(listener)
            input_widget.observe(on_user_change)

    def submit(self):
        def reenable():
            if self.confirm_button is not None:
                self.confirm_button.disabled = False
            self.state.is_submitting = False

        if self.state.is_submitting:
            self.submit_debouncer("", self.submit)
            return self
        self.state.is_submitting = True
        display_value = [input_widget.value for input_widget in self.input_widgets]
        processor = Processor(self.context)
        if self.confirm_button is not None:
            self.confirm_button.disabled = True
        task = processor.create_task(self.input, self.state.runnables, display_value)
        task.add_done_callback(lambda _: reenable())
        return self

    def update_widget(self):
        self.title_widget.value = self.state.title if self.state.title else ""
        self.title_widget.layout.visibility = (
            "visible" if self.state.title else "hidden"
        )

        if self.confirm_button is not None:
            self.confirm_button.description = self.state.action_label
        if self.state.instant_update is True:
            self.confirm_button.disabled = True
            self.confirm_button.layout.visibility = "hidden"
            self.submit_area.layout.visibility = "hidden"
        else:
            self.confirm_button.disabled = False
            self.confirm_button.layout.visibility = "visible"
            self.submit_area.layout.visibility = "visible"

    def __repr__(self):
        return ""

    def show(self):
        main_panel = (
            self.context.main_panel if self.context.main_panel is not None else None
        )
        root_widget = main_panel.widget if main_panel is not None else self.widget
        display(root_widget)
        self.focus()
        return self

    def focus(self):
        if len(self.input_widgets) > 0:
            self.input_widgets[0].focus()
        return self

    def title(self, new_title: str):
        self.state.title = new_title
        self.update_widget()
        return self

    def run(self, *args):
        self.state.runnables = args_to_list(args, defaults=[])
        return self

    def action_label(self, new_label: str):
        self.state.action_label = new_label
        self.update_widget()
        return self

    def instant_update(self, value: bool = True):
        self.state.instant_update = value
        self.update_widget()
        return self

    def save_changes(self, value: bool):
        self.context.save_changes = value
        return self

    def form(self, *args):
        context = self.context
        log_view = self.log_view
        input = args_to_list(args, defaults=[])
        new_form = FormPanel(input=input, context=context, log_view=log_view)

        if len(self.widget.children) == 0:
            self.widget.children = [new_form.widget]
            return new_form

        # Chain the form if necessary
        last_children = self.widget.children[-1]
        if isinstance(last_children, LogViewWidget):
            self.widget.children = tuple(
                list(self.widget.children[:-1]) + [new_form.widget, last_children]
            )
        else:
            self.widget.children = tuple(list(self.widget.children) + [new_form.widget])
        return new_form

    def refresh(self):
        for input_widget in self.input_widgets:
            input_widget.update_widget()
        return self



from jscaffold.patchers.assign import PatchAssignment, QuoteType
import textwrap


def test_assignment_patch_extract_original_value():
    source = textwrap.dedent(
        """\
        a = '10'
        """
    )

    patcher = PatchAssignment()
    _, original_value = patcher(source, "b")

    assert original_value is None

    _, original_value = patcher(source, "a")

    assert original_value == "10"


def test_assignment_patch_shell_script():
    test_cases = [
        (
            textwrap.dedent(
                """\
                a=1
                b=2
                """
            ),
            "a",
            "3",
            textwrap.dedent(
                """\
                a=3
                b=2
                """
            ),
        ),
        (
            textwrap.dedent(
                """\
                a=1
                 b = 2
                """  # noqa
            ),
            "b",
            "10",
            textwrap.dedent(
                """\
                a=1
                 b = 10
                """
            ),
        ),
        (
            textwrap.dedent(
                """\
                a=1
                b ='2'
                """
            ),
            "b",
            "10",
            textwrap.dedent(
                """\
                a=1
                b ='10'
                """
            ),
        ),
        (
            textwrap.dedent(
                """\
                a=1
                export b ='2'
                """
            ),
            "b",
            "10",
            textwrap.dedent(
                """\
                a=1
                export b ='10'
                """
            ),
        ),
        (
            textwrap.dedent(
                """\
                # Comment
                a=1
                export b ='2' # Comment
                """
            ),
            "b",
            "10",
            textwrap.dedent(
                """\
                # Comment
                a=1
                export b ='10' # Comment
                """
            ),
        ),
        (
            # Ignore commented
            textwrap.dedent(
                """\
                # Comment
                a=1
                # b=3
                export b ="2" # Comment
                """
            ),
            "b",
            "10",
            textwrap.dedent(
                """\
                # Comment
                a=1
                # b=3
                export b ="10" # Comment
                """
            ),
        ),
        (
            # Append if not available
            textwrap.dedent(
                """\
                a=1
                b=2
                """
            ),
            "c",
            "3",
            textwrap.dedent(
                """\
                a=1
                b=2

                c=3"""
            ),
        ),
        (
            # Escape double quotes
            textwrap.dedent(
                """\
                a=1.1
                b=2
                """
            ),
            "c",
            '"3"',
            textwrap.dedent(
                """\
                a=1.1
                b=2

                c="\\\"3\\\"\""""
            ),
        ),
        (
            # It should add quote if space existed
            textwrap.dedent(
                """\
                a=1.1
                b=2
                """
            ),
            "a",
            "3 4",
            textwrap.dedent(
                """\
                a="3 4"
                b=2
                """
            ),
        ),
        (
            # It should add quote if new line existed
            textwrap.dedent(
                """\
                a=1.1
                b=2
                """
            ),
            "a",
            textwrap.dedent(
                """\
            3
            4
            """
            ),
            textwrap.dedent(
                """\
                a="3
                4
                "
                b=2
                """
            ),
        ),
    ]

    patcher = PatchAssignment()

    for content, variable, replace, expected in test_cases:
        result, _ = patcher(content, variable, replace)
        assert result == expected


def test_assignment_patch_pod_spec():
    source = textwrap.dedent(
        """\
        Pod::Spec.new do |spec|
        spec.name         = 'Reachability'
        spec.version      = '3.1.0'
        spec.license      = { :type => 'BSD' }
        spec.homepage     = 'https://github.com/tonymillion/Reachability'
        spec.authors      = { 'Tony Million' => 'tonymillion@gmail.com' }
        spec.summary      = 'ARC and GCD Compatible Reachability Class for iOS and OS X.'
        spec.source       = { :git => 'https://github.com/tonymillion/Reachability.git', :tag => 'v3.1.0' }
        spec.source_files = 'Reachability.{h,m}'
        spec.framework    = 'SystemConfiguration'
        end
        """  # noqa
    )

    expected = textwrap.dedent(
        """\
        Pod::Spec.new do |spec|
        spec.name         = 'Reachability'
        spec.version      = '3.1.1'
        spec.license      = { :type => 'BSD' }
        spec.homepage     = 'https://github.com/tonymillion/Reachability'
        spec.authors      = { 'Tony Million' => 'tonymillion@gmail.com' }
        spec.summary      = 'ARC and GCD Compatible Reachability Class for iOS and OS X.'
        spec.source       = { :git => 'https://github.com/tonymillion/Reachability.git', :tag => 'v3.1.0' }
        spec.source_files = 'Reachability.{h,m}'
        spec.framework    = 'SystemConfiguration'
        end
        """  # noqa
    )

    patcher = PatchAssignment()
    result, _ = patcher(source, "spec.version", "3.1.1")
    assert result == expected


def test_assignment_patch_python():
    source = textwrap.dedent(
        """\
        __VERSION__ ='3.1.0'
        """
    )

    expected = textwrap.dedent(
        """\
        __VERSION__ ='3.2.0'
        """
    )

    patcher = PatchAssignment()
    result, _ = patcher(source, "__VERSION__", "3.2.0")
    assert result == expected


def test_assignment_normalize():
    patcher = PatchAssignment()

    assert patcher.normalize("a") == "a"

    assert patcher.normalize("a b") == '"a b"'

    assert patcher.normalize("a", quote=QuoteType.Double) == '"a"'

    assert patcher.normalize("a", quote=QuoteType.Single) == "'a'"
    assert patcher.normalize("'a'", quote=QuoteType.Single) == "'\\'a\\''"



from jscaffold.iounit.var import Var
from jscaffold.utils import inspect_arg_name


def _normalize_input(input):
    if isinstance(input, str):
        return Var(input)
    return input


def _setup_iot(input, output, title):
    if isinstance(input, list):
        input = [_normalize_input(i) for i in input]
    else:
        input = _normalize_input(input)

    return (input, output, title)


def preset_iot(func):
    """
    A decorator that preset the input, output, title arguments
    """

    def inner(input=None, output=None, title=None, *args, **kwargs):
        if title is None:
            if input is not None:
                title = inspect_arg_name(0, "input")
            elif output is not None:
                title = inspect_arg_name(1, "output")
        (input, output, title) = _setup_iot(input, output, title)
        return func(input, output, title, *args, **kwargs)

    return inner


def preset_iot_class_method(func):
    """
    A decorator that preset the input, output, title arguments
    """

    def inner(self, input=None, output=None, title=None, *args, **kwargs):
        if title is None:
            if input is not None:
                title = inspect_arg_name(0, "input")
            elif output is not None:
                title = inspect_arg_name(1, "output")
        (input, output, title) = _setup_iot(input, output, title)
        return func(self, input, output, title, *args, **kwargs)

    return inner



from jscaffold.panel.formpanel import FormPanel
from jscaffold.utils import args_to_list
from jscaffold.views.logview.logview import LogViewWidget
from jscaffold.widgets.ipywidgetwrapper import IPYWidgetWrapper
from ..contexts.context import Context
from ipywidgets import widgets


class MainPanel(IPYWidgetWrapper):
    class State:
        def __init__(self):
            self.forms = []

    def __init__(self):
        self.log_view = LogViewWidget()
        self.context = Context(log_view=self.log_view, main_panel=self)
        self.create_widget()
        self.state = MainPanel.State()

    def create_widget(self):
        self.form_box = widgets.VBox([])
        self.widget = widgets.VBox([self.form_box, self.log_view])

    def update_widget(self):
        self.form_box.children = [form.widget for form in self.state.forms]

    def get_widget(self):
        return self.widget

    def form(self, *args):
        input = args_to_list(args, defaults=[])
        form = FormPanel(input, log_view=self.log_view, context=self.context)
        self.state.forms.append(form)
        self.update_widget()

        return form



from unittest import IsolatedAsyncioTestCase
from jscaffold.iounit.envvar import EnvVar
from jscaffold.panel.formpanel import FormPanel
from jscaffold.services.changedispatcher import ChangeDispatcher
from unittest.mock import Mock
import pytest
import asyncio


class TestChangeDispatcher(IsolatedAsyncioTestCase):
    @pytest.mark.asyncio()
    async def test_listen(self):
        listener = Mock()
        dispatcher = ChangeDispatcher()
        dispatcher.add_listener(listener)
        dispatcher.dispatch("test", "payload")
        await asyncio.sleep(0.2)
        listener.assert_called_with("test", "payload")

    @pytest.mark.asyncio()
    async def test_formpanel_submit(self):
        var1 = EnvVar("VAR1")
        callback = Mock()
        form_panel = FormPanel(var1).run(callback)
        form_panel.submit()
        await asyncio.sleep(0)
        callback.assert_called_once()



import copy
from collections import OrderedDict


class PatchDict:
    def __init__(self):
        pass

    def read(self, object, path):
        keys = path.split(".")
        current = object
        for key in keys:
            if key not in current:
                return None
            if type(current) in [dict, OrderedDict]:
                current = current[key]
            else:
                return None

        return current

    def write(self, object, path, value):
        keys = path.split(".")
        copied = copy.deepcopy(object)
        if type(copied) not in [dict, OrderedDict]:
            copied = {}
        parent = None
        last_key = None
        current = copied
        for key in keys[:-1]:
            if type(current) not in [dict, OrderedDict]:
                current = {}
                parent[last_key] = current
            if key not in current:
                current[key] = {}
            last_key = key
            parent = current
            current = current[key]

        current[keys[-1]] = value
        return copied



from jscaffold.iounit.envvar import EnvVar
import os
import uuid
from unittest.mock import patch, Mock
from unittest import TestCase


@patch("jscaffold.services.changedispatcher.ChangeDispatcher.dispatch", Mock())
class ScaffoldVarTest(TestCase):
    def test_value(self):
        var = EnvVar("VAR1")
        var.write("value")
        assert var.value == "value"

    def test_value_default(self):
        key = str(uuid.uuid4())
        var = EnvVar(key).defaults("default")
        assert var.value == "default"

    def test_write(self):
        """
        It should update the cached value
        """
        key = str(uuid.uuid4())
        var = EnvVar(key).defaults("V1")
        assert var.value == "V1"

        var.write("V2")
        assert var.value == "V2"

    def test_refresh(self):
        var = EnvVar("VAR1")
        os.environ["VAR1"] = "random-value-in-test-value"
        assert var.value == "random-value-in-test-value"

        os.environ["VAR1"] = "random-value-in-test-value2"
        assert var.value == "random-value-in-test-value"

        assert var.refresh().value == "random-value-in-test-value2"

    def test_update(self):
        var = EnvVar("VAR1").refresh()
        os.environ["VAR1"] = str(uuid.uuid4())
        assert var.value != os.environ["VAR1"]
        assert var.update(str(uuid.uuid4())).value == os.environ["VAR1"]

    def test_multiline(self):
        orig = EnvVar("VAR1")
        assert orig.format.type == "text"
        assert orig.format.multiline is False

        fork = orig.multiline()
        assert fork is orig
        assert orig.format.multiline is True
        assert orig.format.multiline is True

    def test_select(self):
        var = EnvVar("VAR1").select("a", "b", "c")
        assert var.format.select == ["a", "b", "c"]

        var = EnvVar("VAR1").select(["a", "b", "c"])
        assert var.format.select == ["a", "b", "c"]



from enum import Enum
import re
from typing import Pattern, Optional, Tuple

# Reference
# https://github.com/theskumar/python-dotenv/blob/main/src/dotenv/parser.py


class QuoteType(Enum):
    No = "no"
    Double = "double"
    Single = "single"


def make_regex(string: str, extra_flags: int = 0) -> Pattern[str]:
    return re.compile(string, re.UNICODE | extra_flags)


_multiline_whitespace = make_regex(r"\s*", extra_flags=re.MULTILINE)
_key = make_regex(
    r"[_a-zA-Z_][a-zA-Z0-9_\.]*"
)  # Support dot for other format like podspec
_equal_sign = make_regex(r"(=[^\S\r\n]*)")
_single_quoted_value = make_regex(r"'((?:\\'|[^'])*)'")
_double_quoted_value = make_regex(r'"((?:\\"|[^"])*)"')
_unquoted_value = make_regex(r"([^\r\n]*)")
_whitespace = make_regex(r"[^\S\r\n]*")
_export = make_regex(r"(?:export[^\S\r\n]+)?")
_comment = make_regex(r"(?:[^\S\r\n]*#[^\r\n]*)?")
_rest_of_line = make_regex(r".*[^\r\n]*(?:\r|\n|\r\n)?")


class PatchAssignment:
    """
    Patch content with assignment expression
    """

    def __init__(
        self,
        equal_sign=_equal_sign,
        multiline_whitespace=_multiline_whitespace,
        key=_key,
        whitespace=_whitespace,
        double_quoted_value=_double_quoted_value,
        single_quoted_value=_single_quoted_value,
        unquoted_value=_unquoted_value,
        export=_export,
        comment=_comment,
        rest_of_line=_rest_of_line,
    ):
        self.equal_sign = equal_sign
        self.multiline_whitespace = multiline_whitespace
        self.key = key
        self.whitespace = whitespace
        self.double_quoted_value = double_quoted_value
        self.single_quoted_value = single_quoted_value
        self.unquoted_value = unquoted_value
        self.export = export
        self.comment = comment
        self.rest_of_line = rest_of_line

    def parse(self, content: str, pattern) -> Tuple[str, str]:
        match = pattern.match(content)
        if match is None:
            raise ValueError("Invalid format")
        extracted = content[match.start() : match.end()]
        remaining = content[match.end() :]
        return extracted, remaining

    def __call__(
        self, content: str, variable, replacement: Optional[str] = None
    ) -> Tuple[str, Optional[str]]:
        remaining = content
        ret = ""
        value = None

        while len(remaining) > 0:
            try:
                extracted, remaining, _value = self.parse_line(
                    remaining, variable, replacement
                )
                if _value is not None:
                    value = _value
                ret += extracted
            except ValueError:
                return content

        if value is None and replacement is not None:
            ret += f"\n{variable}={self.normalize(replacement)}"

        return ret, value

    def parse_line(
        self, content, variable, replacement
    ) -> Tuple[str, str, Optional[str]]:
        remaining = content
        ret = ""
        value = None

        try:
            extracted, remaining = self.parse(remaining, self.multiline_whitespace)
            ret += extracted

            extracted, remaining = self.parse(remaining, self.export)
            ret += extracted

            key, remaining = self.parse(remaining, _key)
            ret += key

            extracted, remaining = self.parse(remaining, self.whitespace)
            ret += extracted

            extracted, remaining = self.parse(remaining, self.equal_sign)
            ret += extracted

            extracted, remaining = self.parse(remaining, self.whitespace)
            ret += extracted

            if key == variable:
                extracted, remaining, value = self.parse_update_value(
                    remaining, replacement
                )
            else:
                extracted, remaining, _ = self.parse_update_value(remaining, None)
            ret += extracted

            extracted, remaining = self.parse(remaining, self.comment)
            ret += extracted

            return ret, remaining, value
        except ValueError:
            extracted, remaining = self.parse(content, self.rest_of_line)
            return extracted, remaining, value

    def parse_update_value(
        self, content, replacement
    ) -> Tuple[str, str, Optional[str]]:
        value = None
        if content.startswith('"'):
            extracted, remaining = self.parse(content, self.double_quoted_value)
            value = extracted[1:-1]
            if replacement is not None:
                return (
                    f"{self.normalize(replacement, quote=QuoteType.Double)}",
                    remaining,
                    value,
                )
        elif content.startswith("'"):
            extracted, remaining = self.parse(content, self.single_quoted_value)
            value = extracted[1:-1]
            if replacement is not None:
                return (
                    f"{self.normalize(replacement, quote=QuoteType.Single)}",
                    remaining,
                    value,
                )
        else:
            extracted, remaining = self.parse(content, self.unquoted_value)
            value = extracted
            if replacement is not None:
                return self.normalize(replacement), remaining, extracted
        return extracted, remaining, value

    def normalize(self, text, quote=None):
        if quote is QuoteType.Single:
            escaped = text.replace("'", r"\'")
            return f"'{escaped}'"
        elif quote is QuoteType.Double:
            escaped = text.replace('"', r"\"")
            return f'"{escaped}"'
        elif any(pattern in text for pattern in [" ", "\n", '"']):
            return self.normalize(text, quote=QuoteType.Double)
        return text



import asyncio
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock, Mock, patch
from jscaffold.iounit.envvar import EnvVar
from jscaffold.panel.formpanel import FormPanel
import pytest


@patch("jscaffold.services.changedispatcher.ChangeDispatcher.dispatch", Mock())
class TestFormPanel(IsolatedAsyncioTestCase):
    def test_form(self):
        """
        FormPanel can be created
        """
        v = EnvVar("test")
        form1 = FormPanel()
        form2 = form1.form(v)
        assert form1.log_view == form2.log_view

    def test_output_only_form(self):
        """
        Try to create a form with only output.
        No error should be raised.
        """
        FormPanel().run("ls")

    @pytest.mark.asyncio()
    async def test_submit_output_only(self):
        callback = MagicMock()
        form = FormPanel().run(callback)
        form.submit()
        await asyncio.sleep(0)
        callback.assert_called_once()

    @pytest.mark.asyncio()
    async def test_submit_should_be_debounced(self):
        """
        Another submit should not work until the first one is done.
        """
        callback = MagicMock()
        form = FormPanel().run(callback)
        form.submit()  # Run immediately
        assert form.state.is_submitting is True
        form.submit()  # Debounced
        form.submit()
        await asyncio.sleep(0)
        callback.assert_called_once()
        await asyncio.sleep(0.2)
        assert callback.call_count == 2



from ipywidgets import widgets
import uuid


class ColumnFlow:
    def __init__(self):
        self.id = uuid.uuid4()
        self.columns = []
        self.hbox = widgets.HBox([])

    @property
    def widget(self):
        return self.hbox

    @property
    def child_count(self):
        return len(self.hbox.children)

    def _create_scroll_to_end_injection(self):
        js = """
        var src = event.target;
        var parent = src.parentNode;
        while (!parent.classList.contains("widget-hbox")) {
            parent = parent.parentNode;
        }
        parent.scrollLeft = parent.scrollWidth;
        """
        escaped_js = js.replace('"', "&quot;")
        div = f"""
        <img src onerror="{escaped_js}"/>
        """
        html = widgets.HTML(div)
        html.layout.display = "none"
        return html

    def _create_container(self, child):
        layout = widgets.Layout(min_width="300px", overflow="none")
        injected = self._create_scroll_to_end_injection()
        container = widgets.VBox([injected, child], layout=layout)
        return container

    def append(self, child):
        container = self._create_container(child)
        self.hbox.children = (*self.hbox.children, container)

    def truncate(self, last):
        self.hbox.children = self.hbox.children[0:last]



import threading
import subprocess
import asyncio
import time
import psutil
import os

SHELL = "/bin/bash"


class RunTask:
    def __init__(self):
        self.script = ""
        self.threads = []

    # pylama:ignore=C901
    def __call__(self, print=print, flush=None, env=None):
        mutex = threading.Lock()
        pending_messages = []
        running = True
        loop = asyncio.get_event_loop()
        future = loop.create_future()

        def worker():
            nonlocal running
            nonlocal pending_messages
            nonlocal loop
            nonlocal future
            process_env = os.environ.copy()
            if env is not None:
                process_env.update(env)
            try:
                process = subprocess.Popen(
                    self.script,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    shell=True,
                    universal_newlines=True,
                    executable=SHELL,
                    env=process_env,
                )
                self.process = process
                for line in process.stdout:
                    mutex.acquire()
                    pending_messages.append(line)
                    mutex.release()
            except Exception as e:
                raise e

            self.exit_code = self.process.poll()
            mutex.acquire()
            running = False
            mutex.release()

        def write_batch(messages):
            for message in messages:
                print(message)
            if flush is not None:
                flush()

        def writer():
            nonlocal running
            nonlocal pending_messages
            nonlocal print
            while running:
                time.sleep(0.1)
                mutex.acquire()
                if len(pending_messages) > 0:
                    loop.call_soon_threadsafe(write_batch, pending_messages.copy())
                    pending_messages = []
                mutex.release()

            loop.call_soon_threadsafe(future.set_result, self.exit_code)

        worker_thread = threading.Thread(target=worker, args=())
        worker_thread.start()

        writer_thread = threading.Thread(target=writer, args=())
        writer_thread.start()
        self.threads = [worker_thread, writer_thread]

        return future

    def kill(self):
        # Kill the process group
        # https://stackoverflow.com/questions/4789837/how-to-terminate-a-python-subprocess-launched-with-shell-true/4791612#4791612
        process = psutil.Process(self.process.pid)
        for proc in process.children(recursive=True):
            proc.kill()
        process.kill()

        for thread in self.threads:
            thread.join()



# pylama:ignore=W0611,W0401
from .envfilevar import *
from .envvar import *
from .jsonfilevar import *
from .applytosource import *
from .format import *
from .var import *
from .blockinfilevar import *



class PatchBlock:
    def __init__(self, begin_divider: str, end_divider: str):
        self.begin_divider = begin_divider
        self.end_divider = end_divider

    def _process(self, content: str):
        lines = content.split("\n")
        begin_index = None
        end_index = None
        for i, line in enumerate(lines):
            if begin_index is None and line.strip() == self.begin_divider:
                begin_index = i
                continue
            if begin_index is not None and line.strip() == self.end_divider:
                end_index = i
                break

        return begin_index, end_index, lines

    def read(self, content: str):
        begin_index, end_index, lines = self._process(content)
        if begin_index is not None and end_index is not None:
            return "\n".join(lines[begin_index + 1 : end_index])

        return None

    def write(self, content: str, value: str):
        begin_index, end_index, lines = self._process(content)

        if begin_index is not None and end_index is not None:
            lines[begin_index + 1 : end_index] = value.split("\n")
            return "\n".join(lines)

        lines = lines + [self.begin_divider] + value.split("\n") + [self.end_divider]
        return "\n".join(lines)



from unittest.mock import MagicMock
import uuid
from jscaffold.iounit.applytosource import ApplyToSource
from jscaffold.iounit.envvar import EnvVar
from jscaffold.processor import Processor
import pytest


@pytest.mark.asyncio()
async def test_should_ignore_readonly():
    context = MagicMock()
    key = str(uuid.uuid4())
    input = EnvVar(key).readonly()
    context.input = input
    processor = Processor(context)
    await processor(input, ApplyToSource(), "123")
    assert input.read() is None



from unittest import TestCase
from unittest.mock import patch
from jscaffold.iounit.envvar import EnvVar
from jscaffold.iounit.format import Formatable, Format, Formatter
import os


class Var(Formatable):
    def __init__(self):
        self.format = Format()


class TestFormatable(TestCase):
    def test_format_select_none(self):
        var = Var()
        var.select("1", "2")
        assert var.format.select == ["1", "2"]
        var.select()
        assert var.format.select is None

    def test_format_defaults_number(self):
        os.environ.pop("VAR", None)
        var = EnvVar("VAR").defaults(1).refresh()
        # It read from defaults and that is why it become 1 instead of "1"
        assert var.value == 1

    @patch("jscaffold.services.changedispatcher.change_dispatcher.dispatch")
    def test_formatable_should_dispatch_change(self, mock_dispatch):
        EnvVar("VAR").select("1", "2")

        mock_dispatch.assert_called_once_with(
            "Env:VAR",
            {"type": "format_changed", "field": "select", "value": ["1", "2"]},
        )


class TestFormatter(TestCase):
    def test_cast_to_format_numer(self):
        var = Var()
        var.number()
        assert Formatter(var.format, "1").cast_to_format().value == 1.0
        assert Formatter(var.format, None).cast_to_format().value is None

    def test_cast_to_format_text(self):
        var = Var()
        assert Formatter(var.format, "1").cast_to_format().value == "1"
        assert Formatter(var.format, None).cast_to_format().value is None
        assert Formatter(var.format, 1.0).cast_to_format().value == "1"



from contextlib import contextmanager
from jscaffold.iounit.valuable import Valuable
from jscaffold.patchers.block import PatchBlock


class BlockInFileVar(Valuable):
    default_filename = None

    def __init__(self, key, filename=None):
        super().__init__()
        self.filename = filename
        if filename is not None:
            self.filename = filename
        elif BlockInFileVar.default_filename is not None:
            self.filename = BlockInFileVar.default_filename
        else:
            raise ValueError("filename is not provided")

        self._key = key
        if isinstance(key, str):
            self.patcher = PatchBlock(key, key)
        else:
            self.patcher = PatchBlock(key[0], key[1])
        self.content = None

    def _get_key(self):
        return self._key

    def _get_id(self):
        return f"BlockInFile:{self.filename}:{self.key}"

    def _write(self, value, context=None):
        content = self._read_file_content()
        replaced = self.patcher.write(content if content is not None else "", value)

        file = open(self.filename, "w")
        file.write(replaced)
        file.close()

        if context is not None and context.log is not None:
            context.log(
                f"Set {self.key}={self._format_display_value(value)} to {self.filename}\n"
            )

    def _read(self, context=None):
        content = self._read_file_content()
        if content is None:
            return None

        value = self.patcher.read(content)
        return value

    def _read_file_content(self):
        try:
            file = open(self.filename, "r")
            content = file.read()
            file.close()
            self.content = content
        except FileNotFoundError:
            return None
        return content

    @classmethod
    @contextmanager
    def use(cls, filename: str):
        try:
            BlockInFileVar.default_filename = filename
            yield
        finally:
            BlockInFileVar.default_filename = None



from typing import Any, Optional
from abc import ABC
from ..contexts.context import Context
from jscaffold.services.changedispatcher import change_dispatcher


class Inputable(ABC):
    """
    Inputable is a class that represents an input unit.
    """

    def __str__(self):
        return self.to_string()

    def to_string(self, context: Context = None) -> str:
        ret = self._read(context=context)
        return ret if ret is not None else ""

    def read(self, context: Context = None) -> Optional[Any]:
        return self._read(context=context)

    def _read(self, context: Context = None) -> Optional[Any]:
        """
        Read the raw value. If the value is not set, return None
        """
        raise NotImplementedError()

    @property
    def id(self):
        return self._get_id()

    @property
    def key(self):
        return self._get_key()

    def _get_key(self):
        raise NotImplementedError()

    def _get_id(self):
        raise NotImplementedError()


class Outputable(ABC):
    def __call__(self, value=None, context: Context = None):
        return self.write(value, context=context)

    def write(self, value=None, context: Context = None):
        ret = self._write(value, context=context)
        if isinstance(self, Inputable):
            object_id = self._get_id()
            payload = {"type": "value_changed", "value": value}
            change_dispatcher.dispatch(object_id, payload)
        return ret

    def _write(self, value=None, context: Context = None):
        raise NotImplementedError()


class IOAble(Inputable, Outputable):
    pass



from enum import Enum
from typing import Any, List, Optional, Union
import re
from jscaffold.iounit.iounit import Inputable
from jscaffold.utils import args_to_list
from jscaffold.services.changedispatcher import change_dispatcher


class FormatType(Enum):
    Text = "text"
    File = "file"
    Number = "number"


class FileSource(Enum):
    Upload = "upload"
    Local = "local"


class FileType(Enum):
    File = "file"
    Directory = "directory"


class Format:
    def __init__(
        self,
        type=FormatType.Text.value,
        defaults: Any = None,
        readonly: bool = False,
        desc: str = None,
        # For text
        multiline: Optional[Union[bool, int]] = False,
        select: Optional[List[str]] = None,
        password: bool = False,
        # For file
        mkdir: bool = True,
        file_source: FileSource = None,
        upload_folder: str = None,
        file_type: FileType = FileType.File,
    ):
        self.type = type
        self.readonly = readonly
        self.defaults = defaults
        self.password = password
        self.desc = desc

        self.multiline = multiline
        self.select = select
        self.file_source = file_source

        # The folder to upload files to, if it is not
        # set, it will use a temp folder
        self.upload_folder = upload_folder
        self.file_type = file_type
        self.mkdir = mkdir


class Formatable:
    def _format_display_value(self, value):
        if self.format.password is True:
            return "*********"
        return value if value is not None else ""

    def _dispatch_format_changed(self, payload: dict):
        if isinstance(self, Inputable):
            object_id = self._get_id()
            change_dispatcher.dispatch(object_id, payload)

    def defaults(self, value):
        self.format.defaults = value
        return self

    def get_defaults(self):
        """
        Get the default value
        """
        defaults = self.format.defaults
        ret = None
        if isinstance(defaults, list):
            ret = defaults[0]
        else:
            ret = defaults

        return ret

    def multiline(self, multiline: Optional[Union[bool, int]] = True):
        self.format.multiline = multiline
        return self

    def select(
        self,
        *args: Optional[List[str]],
    ):
        self.format.select = args_to_list(args, defaults=None)
        self._dispatch_format_changed(
            {"type": "format_changed", "field": "select", "value": self.format.select}
        )
        return self

    def upload_file(self, folder: str = None, mkdir: bool = False):
        self.format.type = FormatType.File.value
        self.format.file_source = FileSource.Upload.value
        self.format.upload_folder = folder
        self.format.mkdir = mkdir
        return self

    def local_path(self, file_type=FileType.File.value):
        self.format.type = FormatType.File.value
        self.format.file_source = FileSource.Local.value
        self.format.file_type = file_type
        return self

    def readonly(self, value=True):
        self.format.readonly = value
        self._dispatch_format_changed(
            {"type": "format_changed", "field": "read_only", "value": value}
        )
        return self

    def password(self, value=True):
        self.format.password = value
        self._dispatch_format_changed(
            {"type": "format_changed", "field": "password", "value": value}
        )
        return self

    def desc(self, value):
        self.format.desc = value
        self._dispatch_format_changed(
            {"type": "format_changed", "field": "read_only", "value": value}
        )
        return self

    def number(self):
        self.format.type = FormatType.Number.value
        self._dispatch_format_changed(
            {"type": "format_changed", "field": "number", "value": None}
        )
        return self


class Formatter:
    def __init__(self, format: Format, value):
        self.format = format
        self.value = value

    def if_none_use_defaults(self):
        defaults = self.format.defaults
        if self.value is not None:
            return self
        if defaults is None:
            return self

        if isinstance(defaults, list):
            self.value = defaults[0]
        else:
            self.value = defaults

        return self

    def cast_to_format(self):
        if self.format.type == FormatType.Number.value:
            self.value = float(self.value) if self.value is not None else None

        if self.format.type == FormatType.Text.value:
            # remove decimal point if it is 0
            if isinstance(self.value, float):
                self.value = str(self.value)
                self.value = re.sub(r"\.0+$", "", self.value)

        return self



from unittest.mock import MagicMock
from unittest import IsolatedAsyncioTestCase
from jscaffold.debounce import KeyFilterDebouncer
import asyncio


class TestKeyFilterDebouncer(IsolatedAsyncioTestCase):
    async def test_rapid_successive_calls(self):
        debouncer = KeyFilterDebouncer(0.1)
        mock = MagicMock()

        def func(value):
            mock(value)

        async def rapid_calls():
            debouncer("start", func, "start")
            for _ in range(5):
                debouncer("rapid", func, "rapid")
                await asyncio.sleep(0)
            await asyncio.sleep(0.2)

        asyncio.create_task(rapid_calls())
        await asyncio.sleep(0.3)
        mock.assert_any_call("rapid")
        mock.assert_any_call("start")
        self.assertEqual(mock.call_count, 2)  # "start" and last "rapid"



from unittest import TestCase
from unittest.mock import Mock, patch
from jscaffold import EnvFileVar
from tempfile import NamedTemporaryFile


@patch("jscaffold.services.changedispatcher.ChangeDispatcher.dispatch", Mock())
class TestEnvFileVar(TestCase):
    def test_envfilevar_get_id(self):
        variable = EnvFileVar("A", "./config.env")
        assert variable._get_id() == "EnvFile:./config.env:A"

    def test_envfile_var_read_not_existed_file(self):
        var = EnvFileVar("var", "not_existed_file").defaults("default")
        assert var._read() is None
        assert var.value == "default"

    def test_envfile_var_write_not_existed_file(self):
        tmp_file = NamedTemporaryFile(delete=True)
        filename = tmp_file.name
        tmp_file.close()

        var = EnvFileVar("A", filename)
        var.write("value")

        file = open(tmp_file.name, "r")
        content = file.read()

        assert content == "\nA=value"

    def test_use(self):
        with EnvFileVar.use("config.env"):
            var1 = EnvFileVar("A")
            assert var1.filename == "config.env"

            var2 = EnvFileVar("B")
            assert var2.filename == "config.env"
        assert EnvFileVar.default_filename is None

    def test_copy(self):
        a = EnvFileVar("A", "config.env")
        a.defaults("default")
        b = a.copy()

        assert b.key == "A"
        assert b.filename == "config.env"
        assert b.format.defaults == "default"

        b.defaults("new_default")
        assert a.format.defaults == "default"



#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function
from glob import glob
import os
from os.path import join as pjoin
from setuptools import find_packages, setup
from jscaffold._version import __version__

HERE = os.path.dirname(os.path.abspath(__file__))

name = "jscaffold"
version = __version__

setup_args = dict(
    name=name,
    description="Jupyter Scaffold",
    version=version,
    scripts=glob(pjoin("scripts", "*")),
    author="Ben lau",
    author_email="xbenlau@gmail.com",
    url="https://github.com/benlau/jscaffold",
    license="BSD",
    platforms="Linux, Mac OS X, Windows",
    keywords=["Jupyter", "Widgets", "IPython"],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Framework :: Jupyter",
    ],
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=["ipywidgets>=7.0.0", "anywidget>=0.9.6"],
    extras_require={
        "test": [
            "pytest>=4.6",
            "pytest-cov",
            "nbval",
        ],
        "examples": [
            # Any requirements for the examples to run
        ],
        "docs": [],
    },
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    entry_points={},
)

if __name__ == "__main__":
    setup(**setup_args)



from jscaffold.patchers.dict import PatchDict
from collections import OrderedDict


def test_dict_read():
    test_cases = [
        ({"a": "b"}, "a", "b"),
        ({"a": {"b": "c"}}, "a.b", "c"),
        ({"a": {}}, "a", {}),
        ({"a": "b"}, "a.b.c", None),
    ]

    for input, path, expected in test_cases:
        patcher = PatchDict()
        actual = patcher.read(input, path)
        assert actual == expected


def test_dict_read_ordered():
    object = OrderedDict({"a": OrderedDict({"b": OrderedDict({"c": "value"})})})
    patcher = PatchDict()
    value = patcher.read(object, "a.b.c")
    assert value == "value"


def test_dict_write():
    test_cases = [
        ({"a": "b"}, "a", "c", {"a": "c"}),
        ({"a": {"b": "c"}}, "a.b", "d", {"a": {"b": "d"}}),
        ({"a": {"b": "c"}}, "a", "d", {"a": "d"}),
        ({"a": "b"}, "a.b.c", "d", {"a": {"b": {"c": "d"}}}),
        ([], "a.b.c", "d", {"a": {"b": {"c": "d"}}}),
        ({"a": []}, "a.b.c", "d", {"a": {"b": {"c": "d"}}}),
        ({}, "a", "c", {"a": "c"}),
        ({}, "a.b", "c", {"a": {"b": "c"}}),
    ]

    for input, path, value, expected in test_cases:
        patcher = PatchDict()
        actual = patcher.write(input, path, value)
        assert actual == expected


def test_dict_write_ordered():
    object = OrderedDict({"a": OrderedDict({"b": OrderedDict({"c": "value"})})})
    patcher = PatchDict()
    actual = patcher.write(object, "a.b.c", "new_value")
    assert type(actual) is OrderedDict
    assert actual == OrderedDict(
        {"a": OrderedDict({"b": OrderedDict({"c": "new_value"})})}
    )



from ipywidgets import widgets
from IPython.display import display
import uuid


# Logger is a view class that is used to log messages
class Logger:
    def __init__(self, line_limit=100):
        self.id = uuid.uuid4()
        self.line_limit = line_limit
        self.line_count = 0
        self.content = ""
        self._create_widget()

    def _create_widget(self):
        html = widgets.HTML(value=self.render())
        vbox = widgets.VBox([html])
        self.widget = vbox
        self.html = html

    def append_stdout(self, message):
        self.content = self.content + message
        self.refresh()

    def clear_output(self):
        self.content = ""
        self.refresh()

    def refresh(self):
        self.html.value = self.render()

    def show(self):
        display(self.widget)

    def render(self):
        normalized_id = str(self.id).replace("-", "")
        container_class_name = f"scaffold_container_{normalized_id}"
        copy_button_class_name = f"scaffold_copy_button_{normalized_id}"

        js = f"""
            var range = document.createRange();
            var elm = document.getElementById('{normalized_id}');
            range.selectNode(elm);
            window.getSelection().removeAllRanges();
            window.getSelection().addRange(range);
            document.execCommand("copy");
            window.getSelection().removeAllRanges();
        """.replace(
            '"', "&quot;"
        )
        # widgets.html can't insert script block, so we use escaped js
        return f"""
<div class="{container_class_name}">
    <style>
    .{container_class_name} {{
        border: 1px solid #bfbfbf; min-height: 80px; overflow-wrap: anywhere;
        width: 100%;
        max-width: 100%;
    }}
    .{copy_button_class_name}:hover {{
        opacity: 0.5;
    }}
    .{copy_button_class_name}:active {{
        opacity: 0.8;
    }}
    .{copy_button_class_name} {{
        position: absolute;
        right: 0px;
        top: 0px;
        border: 1px solid #bfbfbf;
        display: block;
        border: 1px solid #bfbfbf;
        padding: 2px 4px;
        user-select: none;
    }};
    </style>
    <pre id={normalized_id}>{self.content}</pre>
    <div class="{copy_button_class_name}" onclick="{js}">Copy</div>
</div>
"""



from multiprocessing import Process, Pipe

from jscaffold.iounit.format import FileType


def ask_tk_open_file_dialog(file_type: FileType = FileType.File):
    from tkinter import Tk
    from tkinter.filedialog import askopenfilename, askdirectory

    root = Tk()
    root.attributes("-alpha", 0.01)
    root.attributes("-topmost", True)
    root.tk.eval(f"tk::PlaceWindow {root._w} center")
    root.withdraw()

    if file_type == FileType.Directory.value:
        path = askdirectory()
    else:
        path = askopenfilename()
    root.destroy()
    return path


def worker(conn, file_type: FileType = FileType.File):
    try:
        path = ask_tk_open_file_dialog(file_type)
        path = path if path else ""
        conn.send({"path": path})
        conn.close()
    except Exception as e:
        conn.send({"error": str(e)})


class TKService:
    def open_file_dialog(file_type: FileType = FileType.File):
        parent_conn, child_conn = Pipe()
        p = Process(target=worker, args=(child_conn, file_type))
        p.start()
        p.join()
        content = parent_conn.recv()
        if "error" in content:
            raise Exception(content["error"])
        return content["path"]


tk_serivce = TKService



import inspect
import re


def is_varible_name(name: str):
    pattern = re.compile(r"[_a-zA-Z_][a-zA-Z0-9_]*")
    match = pattern.match(name)
    if match is None or match.end() != len(name):
        return False
    return True


def inspect_arg_name(pos: int, name: str):
    try:
        frame = inspect.currentframe()
        frame = inspect.getouterframes(frame)[2]
        frame_string = inspect.getframeinfo(frame[0]).code_context[0].strip()
        args = frame_string[frame_string.find("(") + 1 : -1].split(",")
        args = [arg.strip() for arg in args]
        keyword = {}
        for arg in args:
            token = arg.split("=")
            if len(token) != 2:
                continue
            key = token[0].strip()
            if is_varible_name(key) is False:
                continue
            value = token[1].strip()
            if is_varible_name(value) is False:
                continue
            keyword[key] = value
        if name in keyword:
            return keyword[name]
        else:
            ret = args[pos]
            if is_varible_name(ret) is False:
                return None
            return args[pos]
    except Exception:
        return None


def args_to_list(args, defaults=None):
    if len(args) == 0:
        return defaults

    res = []
    for arg in args:
        if isinstance(arg, list):
            res += arg
        else:
            res.append(arg)
    return res


def link_to_hash(title: str, hash: str):
    return f"<a href='#{hash}'>{title}</a>"
